/* -------------------------------------------------------------------------
   File:        ENG_MAIN.CPP

   Warning:     - All files must be compiled with 4-byte alignment.

   Questions:   - Why do all files need to be compiled with 4-byte
                  alignment?

   DDTK:        Include the following files:
                   loadddtk.cpp
                   ddtk_db.cpp

   BDE32:       Include the following files:
                   newdb.cpp
                   \Borland\BDE\TUTIL32.LIB

                The project path must include:
                   include: $(BCB)\include\vcl

   Copyright (c) 1993-2002 enGenic. All rights reserved.
   ------------------------------------------------------------------------- */
#include "eng_main.h"
#include "tap.h"
#include "initglobals.h"
#include "wait.h"
#include "..\escutil\esc_except.h"
#include "..\escutil\odbc_db.h"
#include "..\xmlutil\xmllist.h"
#include "..\TFServer\TFServer_communication.h"
#include "LoadLanguage.h"
#include "VoiceCreator.h"
#include "..\TFServer\TCPClientManager.h"

/* -------------------------------------------------------------------------
   Name:        EngineMain

   Purpose:     Starts an instance of the Engine running.

   Notes:       - If there is a serious problem starting the Engine, the
                  primary thread will not be informed that the Engine didn't
                  start.  Therefore this thread will not try and start
                  again.
   ------------------------------------------------------------------------- */
TFServerCommunicationPacket::TF_SERVER_EVENT _stdcall EngineMain(void* params)
{
   PIVRINFO pInfo = (PIVRINFO) params;
   IV_STATUS           status;
   TAP                *tap = 0;
   int                 i;
   EXCEPTION_POINTERS *xp;
   string languageDLL;
   bool firstTime = true;

   if((pInfo == NULL) || (pInfo->ivBoard == NULL) || (pInfo->ivLog == NULL))
   {
      // Unable to start Engine thread.  Don't inform server since this
      // the Server simply try and restart it.
      return(TFServerCommunicationPacket::SRV_EVT_ENG_FATAL_ERROR);
   }

   // Init pointers
   pInfo->languageLibrary = NULL;
   pInfo->language = NULL;

	// Load default language module (English)...

   // Set language module filename
   languageDLL = pInfo->pszInstallationPath;
   languageDLL += "Lang_English.dll";

   // Load default language module
   LoadLanguageModule(pInfo, languageDLL.c_str());

   pInfo->loadingMutex->Get();
   tap = new TAP(pInfo);
   pInfo->loadingMutex->Release();

   while(pInfo->bRestart)
   {
      // Continuously reload Engine until shutdown.
      try                 // Structured exception handling under Win32
      {
         try              // Structured exception handling under Win32
         {
            // Init pointers
            pInfo->speechRecognitionLibrary = NULL;
            pInfo->speechRecognition = NULL;
            pInfo->voiceAuthenticationLibrary = NULL;
            pInfo->voiceAuthentication = NULL;

            status = IVR_SUCCESS;

            if(pInfo->restartCnt > 0)
            {
               pInfo->ivLog->Print("Application restart %d\n\n\n", pInfo->restartCnt);
            }

   			if(pInfo->status == lsCrashed)
            {
            	// Status: crashed but restarted and running normally
   	         pInfo->status = lsWarning;
            }
            else
            {
   	         pInfo->status = lsNormal;
            }

            // Reset the log for each iteration except first
            if(!firstTime)
            {
               pInfo->ivLog->Reset();
               firstTime = false;
            }


            // Clear variables
            pInfo->variables->Clear();
            // Clear TeleFlow handles
            pInfo->odbcManager->Reset();
            pInfo->xmlList->Reset();

            tap->SetSysVariables(pInfo);

   			// Create system variables for parameters (@P1 - @P9)
            for (i = 0; i < pInfo->iParamCount; i++) {
               char  parmNm[8];
               sprintf (parmNm, "@P%d", i + 1);
               // Parameters from the command line are always global variables.
               pInfo->variables->SetSystem(parmNm, pInfo->pszParams[i]);
            }

            InitializeGlobals(pInfo);

            pInfo->reloadRequested = false;
            pInfo->shutdownRequested = false;
            pInfo->safeToShutdown = false;

            pInfo->ivLog->Debug ('p', "Initializing...\n\n");
            pInfo->ivBoard->Reset();

            if(status == IVR_SUCCESS)
            {
               pInfo->ivLog->Print("\nInitialization done\n\n");

               if(tap->GetReady())
               {
                  status = tap->Execute(pInfo);
               }
               else
               {
                  // TAP not ready when something in the loading process fails (eg: missing file)
                  status = IVR_FAIL;
               }
            }

            pInfo->ivBoard->Hangup();
            pInfo->ivBoard->RemoveStreamers();

            // Un-initialize COM library when TAP execution ends - TF Jun.02/02
            CoUninitialize();

            // pInfo->status will be lsCrashed if we have crashed even if
            // status is currently good because we crashed nicely such as the FATAL
            // handler
            if((pInfo->status != lsCrashed) && ((status == IVR_SUCCESS)
            	|| (status == IVR_SHUTDOWN) || (status == IVR_ENDPROC)))
   			{
               pInfo->status = lsNotRunning;
            }
            else
            {
               pInfo->status = lsCrashed;
            }
         }
         __except (xp = GetExceptionInformation(), EXCEPTION_EXECUTE_HANDLER)
         {
   			unsigned int      xc;
            EXCEPTION_RECORD *xr;

   			xc = GetExceptionCode();
            xr = xp->ExceptionRecord;

   			switch (xc)
            {
   	         case EXCEPTION_ACCESS_VIOLATION:
 	            {
 						pInfo->ivLog->Print("%lX: EXCEPTION_ACCESS_VIOLATION attempting to "
                              " %s at location %lX.\n",
                              xr->ExceptionAddress,
                              (xr->ExceptionInformation[0] ? "write" : "read"),
                              xr->ExceptionInformation[1]);
		            pInfo->ivLog->Print("Exception occurred!  Thread exiting.\n");

                  break;
               }

               default:
               {
 						pInfo->ivLog->Print("EXCEPTION CODE %x.\n", xc);
  		            pInfo->ivLog->Print("Exception occurred!  Thread exiting.\n");
               }
            } // end switch

            pInfo->status = lsCrashed;
         } // end __except
      }
      __finally
      {
         // Free speech recognition memory and release licenses
         if(pInfo->speechRecognitionLibrary != NULL)
         {
            PVoiceDestroyer voiceDestroyer = GetVoiceDestroyer(pInfo->speechRecognitionLibrary);

            if(voiceDestroyer != NULL)
            {
               voiceDestroyer(pInfo->speechRecognition);
            }

            pInfo->speechRecognition = NULL;
            pInfo->speechRecognitionLibrary = NULL;
         }

         // Free voice authentication memory and release licenses
         if(pInfo->speechRecognitionLibrary != NULL)
         {
            PVoiceDestroyer voiceDestroyer = GetVoiceDestroyer(pInfo->voiceAuthenticationLibrary);

            if(voiceDestroyer != NULL)
            {
               voiceDestroyer(pInfo->voiceAuthentication);
            }

            pInfo->voiceAuthentication = NULL;
            pInfo->voiceAuthenticationLibrary = NULL;
         }

         // Process restart
         if(pInfo->bTerminate || pInfo->shutdownRequested
            || (!pInfo->bRestartInfinite && (pInfo->restartCnt >= pInfo->restartMax)))
         {
            pInfo->bRestart = false;
         }
         else
         {
            pInfo->bRestart = true;
            pInfo->restartCnt++;
         }

         // Reset information for requested restart
         if(pInfo->shutdownRequested)
         {
            pInfo->status = lsNotRunning;
            pInfo->restartCnt = 0;
         }
         // THE FOLLOWING VALUE SHOULD BE SOFT-CODED INSTEAD - TF Jul.26/02
         // Pause on exit to prevent auto-reload from choking the CPU - TF Jul.26/02
         else if(!pInfo->bTerminate)
         {
            Wait(pInfo, pInfo->restartDelay);
         }

         // Nice log spacing
         pInfo->ivLog->Print("\n\n\n");

   		// We're supposed to log on errors
         if (!pInfo->bLogOutput && (pInfo->status == lsCrashed))
         {
            pInfo->ivLog->WriteToFile(((string) pInfo->pszLogDir + (string) pInfo->pszAppNm).c_str(), true);
         }

         pInfo->tcpClientManager->UpdateAppStatus(pInfo->iAppID, pInfo->status);
      }
   }

   // Free memory allocated in this thread.
   delete tap;
   UnloadLanguageModule(pInfo);

   pInfo->ivLog->Print("All memory released.  Exiting...\n\n\n");

   return(TFServerCommunicationPacket::SRV_EVT_ENG_THRD_DONE);
}
/*EngineMain*/
