#include "step_fileopen.h"

StepFileOpen::StepFileOpen(STEPREC& stepRecord) : Step(stepRecord)
{
   fileNm = stepRecord.s6.arr;
   varNm = stepRecord.s1.arr;
   varResultByteCount = stepRecord.s2.arr;
   varBaseByteCount = stepRecord.s3.arr;
   bBase64Encode = stepRecord.u1;
   bBreakLines = stepRecord.u3;
   bBase64CountCRLF = stepRecord.u2;
}

StepFileOpen::~StepFileOpen(void)
{
}

const char* StepFileOpen::GetStepName(void) const
{
	return("Load File");
}

IV_STATUS StepFileOpen::Execute(PIVRINFO pInfo)
{
   IV_STATUS status;
   char* value;
   char* pathName;
   char formatstr[24];
   unsigned int iResultCount;
   unsigned int iUnconvertedCount;
   bool bParmsValid = true;   // to evaluate all parameters before exiting

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   if((value = pInfo->variables->ParseParameter("File", fileNm.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      pathName = pInfo->variables->FilePath(value);
      delete[] value;
   }

   if(bBase64Encode)
   {
      pInfo->ivLog->Print("      Encode to Base64\n");
   }

   if(!pInfo->variables->IsVariableFormat(varNm.c_str()))
   {
      pInfo->ivLog->Print("      ERROR: Invalid variable name '%s'\n", varNm.c_str());
      bParmsValid = false;
   }

   if((!pInfo->variables->IsEmpty(varResultByteCount.c_str())) &&
      (!pInfo->variables->IsVariableFormat(varResultByteCount.c_str())))
   {
      pInfo->ivLog->Print("      ERROR: Invalid variable name '%s'\n", varResultByteCount.c_str());
      bParmsValid = false;
   }

   if((!pInfo->variables->IsEmpty(varBaseByteCount.c_str())) &&
      (!pInfo->variables->IsVariableFormat(varBaseByteCount.c_str())))
   {
      pInfo->ivLog->Print("      ERROR: Invalid variable name '%s'\n", varBaseByteCount.c_str());
      bParmsValid = false;
   }

   if(bParmsValid == false)
   {
      return(IVR_FATAL);
   }

   status = pInfo->variables->LoadFile(varNm.c_str(),
                                       pathName,
                                       iUnconvertedCount,
                                       iResultCount,
                                       bBase64Encode,
                                       bBreakLines,
                                       bBase64CountCRLF);

   if(status == IVR_SUCCESS)
   {
      if(!pInfo->variables->IsEmpty(varBaseByteCount.c_str()))
      {
         sprintf(formatstr, "%i", iUnconvertedCount);
         if(!pInfo->variables->Set(varBaseByteCount.c_str(), formatstr))
         {
            status = IVR_FATAL;
         }
      }

      if(!pInfo->variables->IsEmpty(varResultByteCount.c_str()))
      {
         sprintf(formatstr, "%i", iResultCount);
         if(!pInfo->variables->Set(varResultByteCount.c_str(), formatstr))
         {
            status = IVR_FATAL;
         }
      }
   }

   delete[] pathName;

   return(status);
}
