//---------------------------------------------------------------------------

// Must be included before vcl.h so that winsock2.h doesn't interfere with windows.h
#include "..\inet\Socket.h"
// Must be included before vcl.h so that Borland doesn't interfere with the std::map code in SignalQueue
#include "..\TFServer\signalevt.h"

#include <vcl.h>
#pragma hdrstop

#include <mmsystem.h>                        	// MCI stuff
#include <process.h>                            // _beginthreadex
#include <dos.h>                                // _argc, _argv.
#include <dir.h>                                // chdir
#include <jpeg.hpp>                             // JPEG support
#include "..\TFServer\nt_service.h"             // TFSERVER_REG_KEY
#include "..\TFServer\brd_load.h"               // LoadBoardFunctions()
#include "..\TFEngine\debugopt.h"               // EngineMain()
#include "..\escutil\help_fns.h"                // for link to Help file entry
#include "TFSimulator_communication.h"          // TFSimulatorCommunicationPacket
#include "msg_win.h"                            // TMsgWindow
#include "TFSimulator_about.h"                  // TAboutBox
#include "TFSimulator_logdir.h"                 // TLogDirForm
#include "TFSimulator_logsize.h"                // TLogSizeForm
#include "TFSimulator_placecall.h"              // TPlaceCallForm
#include "..\TFServer\File_Log.h"
#include "TFSimulator_main.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFormSimulator *FormSimulator;
//---------------------------------------------------------------------------

// Thread to receive all messages for the simulator or server and process them
static unsigned int _stdcall MessageMonitor(void* params)
{
   // Pointer to simulator object for method calls in response to received messages
   TFormSimulator* simulator = ((MonitorParameters*) params)->simulator;

   // Message queue for messages to simulator
   TFSimulatorCommunicationDestination* simulatorMessages = ((MonitorParameters*) params)->simulatorMessages;

   // Message queue for messages to server
   TFServerCommunicationDestination* serverMessages = ((MonitorParameters*) params)->serverMessages;

   // Array of 3 handles for shutting down this thread or receiving messages from the queues
   HANDLE* waitObjs = ((MonitorParameters*) params)->waitObjs;

   for(DWORD ret = WaitForMultipleObjects(3, waitObjs, false, INFINITE);;
      ret = WaitForMultipleObjects(3, waitObjs, false, INFINITE))
   {
      switch(ret)
      {
         // Service is shutting down
         case WAIT_OBJECT_0:
         {
            return(0);
         }

         // New Simulator message
         case WAIT_OBJECT_0 + 1:
         {
            TFSimulatorCommunicationPacket* message = (TFSimulatorCommunicationPacket*) simulatorMessages->ReceiveMessage();

            switch(message->event)
            {
               case TFSimulatorCommunicationPacket::TF_SIMULATOR_ON_HOOK:
               {
                  simulator->ProcessOnHook();
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_OFF_HOOK:
               {
                  simulator->ProcessOffHook();
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_READY_FOR_CALL:
               {
                  simulator->ProcessReadyForCall(true);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_NOT_READY_FOR_CALL:
               {
                  simulator->ProcessReadyForCall(false);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_REFRESH:
               {
                  simulator->ProcessRefresh(message->status, message->stepNumber, message->stepName, message->TAMName);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_OUTGOING_CALL:
               {
                  simulator->ProcessOutgoingCall();
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_SET_HWI:
               {
                  simulator->SetHWI((HWAVEIN) message->eventInfo);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_SET_MCI_DEVICE:
               {
                  simulator->SetMCIDevice((LONG) message->eventInfo);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_SHOW_STOP_RECORDING_BUTTON:
               {
                  simulator->ShowStopRecordingButton(true);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_HIDE_STOP_RECORDING_BUTTON:
               {
                  simulator->ShowStopRecordingButton(false);
                  break;
               }
/*
               case TFSimulatorCommunicationPacket::TF_SIMULATOR_SHOW_BARGEIN_BUTTON:
               {
                  simulator->ShowBargeInButton(true);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_HIDE_BARGEIN_BUTTON:
               {
                  simulator->ShowBargeInButton(false);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_ENABLE_BARGEIN_BUTTON:
               {
                  simulator->EnableBargeInButton(true);
                  break;
               }

               case TFSimulatorCommunicationPacket::TF_SIMULATOR_DISABLE_BARGEIN_BUTTON:
               {
                  simulator->EnableBargeInButton(false);
                  break;
               }
*/
               default:
               {
               }
            }

            delete message;

            break;
         }

         // New Server message
         case WAIT_OBJECT_0 + 2:
         {
            TFServerCommunicationPacket* message = (TFServerCommunicationPacket*) serverMessages->ReceiveMessage();

            switch(message->event)
            {
               case TFServerCommunicationPacket::SRV_EVT_ENG_THRD_DONE:
               case TFServerCommunicationPacket::SRV_EVT_ENG_FATAL_ERROR:
               case TFServerCommunicationPacket::SRV_EVT_ENG_TERMINATED:
               {
                  simulator->TerminateEngine(message->appID);
                  break;
               }

               default:
               {
                  // Ignore
               }
            }

            delete message;

            break;
         }
      }
   }
};
//---------------------------------------------------------------------------

// Thread to call TFEngine by sending ring events and making ring noises
static unsigned int _stdcall CallSystem(void* params)
{
   TFormSimulator* simulator = ((CallSystemParameters*) params)->simulator;
   HMODULE library = ((CallSystemParameters*) params)->library;
   HANDLE shutdown = ((CallSystemParameters*) params)->shutdown;

   PlaySound("DIAL", library, SND_RESOURCE);

   // Ring until the phone picks up
   do
   {
      // The SIMEVT_RING needs to be added before we play the file so that we cannot
      // hangup before the SIMEVT_RING occurs but pressing the "Call System"/"Hangup"
      // twice quickly in succession.
      simulator->SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_RING);
      PlaySound("RING", library, SND_RESOURCE);
   } while(WaitForSingleObject(shutdown, 3000) == WAIT_TIMEOUT);

   return(0);
}
//---------------------------------------------------------------------------

__fastcall TFormSimulator::TFormSimulator(TComponent* Owner)
   : TForm(Owner), callSystemID(0), onHook(true), testcounter(0)
{
   // Shutdown status monitor thread
   monitorParams.waitObjs[0] = CreateEvent(NULL, true, false, NULL);
   monitorParams.waitObjs[1] = TFSimulatorMessages.GetMessageSignal();
   monitorParams.waitObjs[2] = TFServerMessages.GetMessageSignal();
   monitorParams.simulator = this;
   monitorParams.simulatorMessages = &TFSimulatorMessages;
   monitorParams.serverMessages = &TFServerMessages;

   callSystemParams.simulator = this;
   callSystemParams.library = HInstance;
   callSystemParams.shutdown = CreateEvent(NULL, true, false, NULL);   // Shutdown call system thread

   SimulationMenuPause->Enabled = true;
   lastStatus = lsEmpty;
   currentApplication = NULL;
   currentAppID = 1000;
   hResume = CreateEvent(NULL, true, true, NULL);

   // Add message queues to communication hub
   commHub.Add(&TFServerMessages);
   commHub.Add(&TFSimulatorMessages);

   HelpFile = regInfo.pszInstallationPath;
   HelpFile += "TeleFlow.chm";
}
//---------------------------------------------------------------------------

__fastcall TFormSimulator::~TFormSimulator(void)
{
   CloseHandle(callSystemParams.shutdown);
   CloseHandle(monitorParams.waitObjs[0]);
   CloseHandle(hResume);
}

void __fastcall TFormSimulator::Setup(void)
{
   int ret;
   unsigned threadId;
   HANDLE hMonThread;

   regInfo.OpenSystemLog(sysLog);
   regInfo.Load(sysLog);

   delete[] regInfo.pszBoardDLL;
   regInfo.pszBoardDLL = new char[strlen(regInfo.pszInstallationPath) + strlen("brd_simulator.dll") + 1];
   sprintf(regInfo.pszBoardDLL, "%sbrd_simulator.dll", regInfo.pszInstallationPath);

   ReadTFSimulatorRegistry();

   sysLog.Print("Dynamically loading board DLL at run-time.\n");

   if((boardLibrary = LoadBoardFunctions(&sysLog, regInfo.pszBoardDLL)) == NULL)
   {
      sysLog.Print("ERROR: Unable to load board: %d\n", TFS_BOARD_DLL_PROBLEM);
   }

   sysLog.Print("Board DLL: %s\n", regInfo.pszBoardDLL);
   if(OpenHardwareDriver != NULL)
   {
   	if(OpenHardwareDriver(&sysLog) != true)
      {
			sysLog.Print("ERROR: Could not start telephony hardware driver!\n");
      }
   }

   if((ret = Socket::NetStartup()) != Socket::SUCCESS)
   {
      sysLog.Print("ERROR: Could not start Winsock: '%s'\n", Socket::WSAErrorString(ret));
   }

   tcpClientManager = new TCPClientManager((unsigned short) regInfo.dwMonitorPort,
                                           (unsigned short) regInfo.dwMonitorHTTPPort,
                                           commHub,
                                           &regInfo);

   commHub.Add(tcpClientManager);

   hMonThread = (HANDLE) _beginthreadex (
      NULL,               // NT security attributes
      0,                  // stack size (0 = use same size as this thread)
      TCPClientManagerMain,  // start address
      tcpClientManager,           // argument list
      0,                  // Thread creation flags
      &threadId);         // Thread id
   CloseHandle (hMonThread);

   sysLog.Print ("Creating application manager\n");
   applicationsManager = new TFApplicationsManager(&commHub, tcpClientManager, sysLog, boardLibrary);
   commHub.Add(applicationsManager);

   InitialiseGraphics();

   _beginthreadex(
      NULL,               // NT security attributes
      0,                  // stack size (0 = use same size as this thread)
      MessageMonitor,      // start address
      &monitorParams,     // argument list
      0,                  // Thread creation flags
      &threadId);         // Thread id

   // PlaceCallForm must be shown and then hidden or else it does not display
   // properly later on.  Why is this kludge necessary?
   PlaceCallForm->Show();
   PlaceCallForm->Hide();

   // Fills in simEngineParams.info.pszAppNm, simEngineParams.info.pszAppFile,
   // and simEngineParams.info.pszParams[] from command line parameters
   ReadInParameters();
}

void __fastcall TFormSimulator::Cleanup(void)
{
   WriteRegistry();

   SetEvent(callSystemParams.shutdown);
   SetEvent(monitorParams.waitObjs[0]);
   if(currentApplication != NULL)
   {
      SetEvent(currentApplication->info.hResume);
   }

  	FreeLibrary(callSystemParams.library);

   UnloadBoardFunctions();

   delete applicationsManager;
   delete tcpClientManager;

   Socket::NetShutdown();
}

void __fastcall TFormSimulator::LoadJPGFromResource(TImage* image, const char* jpgName)
{
   TResourceStream* JPGStream = new TResourceStream((int) HInstance, jpgName, "JPGIMAGE");
   TJPEGImage* JPGImage = new TJPEGImage();
   JPGImage->LoadFromStream(JPGStream);
   image->Picture->Bitmap->Assign(JPGImage);
   delete JPGImage;
   delete JPGStream;
}

void __fastcall TFormSimulator::LoadGraphics(void)
{
   // Keypad
   LoadJPGFromResource(KeypadPanelImage, "Panel1");
   LoadJPGFromResource(Button1Image, "tt1u");
   LoadJPGFromResource(Button1DownImage, "tt1p");
   LoadJPGFromResource(Button2Image, "tt2u");
   LoadJPGFromResource(Button2DownImage, "tt2p");
   LoadJPGFromResource(Button3Image, "tt3u");
   LoadJPGFromResource(Button3DownImage, "tt3p");
   LoadJPGFromResource(Button4Image, "tt4u");
   LoadJPGFromResource(Button4DownImage, "tt4p");
   LoadJPGFromResource(Button5Image, "tt5u");
   LoadJPGFromResource(Button5DownImage, "tt5p");
   LoadJPGFromResource(Button6Image, "tt6u");
   LoadJPGFromResource(Button6DownImage, "tt6p");
   LoadJPGFromResource(Button7Image, "tt7u");
   LoadJPGFromResource(Button7DownImage, "tt7p");
   LoadJPGFromResource(Button8Image, "tt8u");
   LoadJPGFromResource(Button8DownImage, "tt8p");
   LoadJPGFromResource(Button9Image, "tt9u");
   LoadJPGFromResource(Button9DownImage, "tt9p");
   LoadJPGFromResource(Button0Image, "tt0u");
   LoadJPGFromResource(Button0DownImage, "tt0p");
   LoadJPGFromResource(ButtonStarImage, "ttsu");
   LoadJPGFromResource(ButtonStarDownImage, "ttsp");
   LoadJPGFromResource(ButtonPoundImage, "ttpu");
   LoadJPGFromResource(ButtonPoundDownImage, "ttpp");

   // Status lights
   NormalStatusLightImage->Picture->Bitmap->LoadFromResourceName((int) HInstance, "LightG");
   CrashedStatusLightImage->Picture->Bitmap->LoadFromResourceName((int) HInstance, "LightR");
   NotRunningStatusLightImage->Picture->Bitmap->LoadFromResourceName((int) HInstance, "LightDG");

   // Phone
   LoadJPGFromResource(OnHookImage, "TeleOnHook");
   LoadJPGFromResource(OffHookImage, "TeleOffHook");

   // Button Panel
   LoadJPGFromResource(ButtonPanelImage, "Panel2");

   // Call System/Hangup Button
   LoadJPGFromResource(CallSystemImage, "ButtCallU");
   LoadJPGFromResource(CallSystemDownImage, "ButtCallP");
   LoadJPGFromResource(CallSystemDisabledImage, "ButtCallD");
   LoadJPGFromResource(HangupImage, "ButtHangupU");
   LoadJPGFromResource(HangupDownImage, "ButtHangupP");
   LoadJPGFromResource(HangupDisabledImage, "ButtHangupD");

   // Pause/Resume Button
   LoadJPGFromResource(PauseImage, "ButtPauseU");
   LoadJPGFromResource(PauseDownImage, "ButtPauseP");
   LoadJPGFromResource(PauseDisabledImage, "ButtPauseD");
   LoadJPGFromResource(ResumeImage, "ButtResumeU");
   LoadJPGFromResource(ResumeDownImage, "ButtResumeP");
   LoadJPGFromResource(ResumeDisabledImage, "ButtResumeD");

   // Next Step Button
   LoadJPGFromResource(NextStepImage, "ButtNextStepU");
   LoadJPGFromResource(NextStepDownImage, "ButtNextStepP");
   LoadJPGFromResource(NextStepDisabledImage, "ButtNextStepD");
   LoadJPGFromResource(NextStepBlankImage, "ButtBlank");

   // Stop Recording Button
   LoadJPGFromResource(StopRecordingImage, "ButtStopRecU");
   LoadJPGFromResource(StopRecordingDownImage, "ButtStopRecP");
   LoadJPGFromResource(StopRecordingDisabledImage, "ButtStopRecD");
   LoadJPGFromResource(StopRecordingBlankImage, "ButtBlank");

   // Show Log Button
   LoadJPGFromResource(ShowLogImage, "ButtShowlogU");
   LoadJPGFromResource(ShowLogDownImage, "ButtShowlogP");

   // Barge-In Button
   LoadJPGFromResource(BargeInImage, "ButtBargeInU");
   LoadJPGFromResource(BargeInDownImage, "ButtBargeInP");
   LoadJPGFromResource(BargeInDisabledImage, "ButtBargeInD");
}

void __fastcall TFormSimulator::ResetGraphics(void)
{
   // Phone
   OnHookImage->Show();
   OffHookImage->Hide();

   // Call System/Hangup Button
   CallSystemImage->Hide();
   CallSystemDownImage->Hide();
   CallSystemDisabledImage->Show();
   HangupImage->Hide();
   HangupDownImage->Hide();
   HangupDisabledImage->Hide();

   // Stop Recording Button
   StopRecordingImage->Hide();
   StopRecordingDownImage->Hide();
   StopRecordingDisabledImage->Hide();
   StopRecordingBlankImage->Show();

   // Barge-In Button
   BargeInImage->Hide();
   BargeInDownImage->Hide();
   BargeInDisabledImage->Hide();
}

void __fastcall TFormSimulator::InitialiseGraphics(void)
{
   LoadGraphics();
   ResetGraphics();

   // Status lights
   NormalStatusLightImage->Hide();
   CrashedStatusLightImage->Hide();
   NotRunningStatusLightImage->Hide();

   // Pause/Resume Button
   PauseImage->Show();
   PauseDownImage->Hide();
   PauseDisabledImage->Hide();
   ResumeImage->Hide();
   ResumeDownImage->Hide();
   ResumeDisabledImage->Hide();

   // Next Step Button
   NextStepImage->Hide();
   NextStepDownImage->Hide();
   NextStepDisabledImage->Hide();
   NextStepBlankImage->Show();

   // Show Log Button
   ShowLogImage->Show();
   ShowLogDownImage->Hide();
}

void __fastcall TFormSimulator::PrepareNewApplication(void)
{
   string logPrefix = (string) regInfo.pszLogDir + (string) "\\TFSimulator";
   string logName = logPrefix;
   char number[14];
   int port = 1;
   SimLog* ivLog = new SimLog(this, regInfo.dwMaxLogFileSize, regInfo.dwLogLines);

   if(currentApplication != NULL)
   {
      ((SimLog*) currentApplication->info.ivLog)->Deactivate();
   }

   // Try to delete file to see if it's in use and increment port if it is and try again
   while((remove(logName.c_str()) == -1) && (errno == EACCES))
   {
      sprintf(number, " %d", ++port);
      logName = logPrefix + (string) number;
   }

   ivLog->Open(logName.c_str(), false, true);
   ivLog->SetDebug('A');
   ivLog->SetDebug(1);  // Allow "true" to be passed to Debug.
   SetupDebugOptions(ivLog);

   // Server level manipulation of the info structure
   currentApplication = new TFApplication(++currentAppID, ivLog);

   currentApplication->info.pszInstallationPath = new char[strlen(regInfo.pszInstallationPath) + 1];
   strcpy(currentApplication->info.pszInstallationPath, regInfo.pszInstallationPath);
   currentApplication->info.pszAppNm = new char[strlen("Simulator Test") + 1];
   strcpy(currentApplication->info.pszAppNm, "Simulator Test");
   currentApplication->info.pszProtocol = new char[strlen("lps0") + 1];
   strcpy(currentApplication->info.pszProtocol, "lps0");
   currentApplication->info.iPort = port;
   currentApplication->info.pszLogDir = new char[strlen(regInfo.pszLogDir) + 1];
   strcpy(currentApplication->info.pszLogDir, regInfo.pszLogDir);
   currentApplication->info.portLicense = LICENSED;
   currentApplication->info.restartMax = 0;
   currentApplication->info.restartCnt = 0;
   currentApplication->info.bRestartInfinite = false;
   currentApplication->info.bLogOutput = true;
   currentApplication->info.bLogAppend = false;
   currentApplication->info.logSRAudio = (bool) regInfo.dwLogSRAudio;
   currentApplication->info.loadingMutex = &loadingMutex;
   currentApplication->info.hSimulator = this->Handle;
   currentApplication->info.hResume = hResume;
   currentApplication->info.commHub = &commHub;
   currentApplication->info.messages = new TFEngineCommunicationDestination(currentAppID);
   applicationsManager->Add(currentApplication->info.messages);
   currentApplication->info.tcpClientManager = tcpClientManager;
   currentApplication->info.boardLibrary = boardLibrary;
   if(CreateInstance != NULL)
   {
      currentApplication->info.ivBoard = CreateInstance(&currentApplication->info);
   }

   applicationsManager->AddApplication(currentAppID, currentApplication);
}

void __fastcall TFormSimulator::ReadInParameters(void)
{
   // Command line parameter offset.
   int index = 1;
   int paramOffset = 0;

   if(_argc <= 1)
   {
      return;
   }

   PrepareNewApplication();

   while(index < _argc)
   {
      if(_argv[index][0] == '-' || _argv[index][0] == '/')
      {
         switch(_argv[index][1])
         {
            // Doesn't account for whitespace.
            case 'D':
            {
               chdir(&_argv[index][2]);
               break;
            }

            // Debug parameters (ignored).
            case 'Z':
            // Not allowing logging.
            case 'F':
            // Ignore unsupported parameter.
            default:
               break;
         }
      }
      else
      {
         switch(++paramOffset)
         {
            // 1st parameter: AppID (is no longer needed)
            case 1:
            {
               break;
            }

            // 2nd parameter: AppName (is no longer needed)
            case 2:
            {
               break;
            }

            // 3rd parameter: App Filename
            case 3:
            {
               currentApplication->info.pszAppFile = new char[strlen(_argv[index]) + 1];
               strcpy(currentApplication->info.pszAppFile, _argv[index]);
               break;
            }

            // 4th to (MAX_PARAMETERS + 3)th parameters: Optional parameters
            default:
            {
               currentApplication->info.pszParams[paramOffset - 4] = new char[strlen(_argv[index]) + 1];
               strcpy(currentApplication->info.pszParams[paramOffset - 4], _argv[index]);
               currentApplication->info.iParamCount++;
            }
         }
      }

      index++;
   } //while

   // Make all @P variables exist
   for(; currentApplication->info.iParamCount < MAX_PARAMETERS; currentApplication->info.iParamCount++)
   {
      currentApplication->info.pszParams[currentApplication->info.iParamCount] = new char[1];
      strcpy(currentApplication->info.pszParams[currentApplication->info.iParamCount], "");
   }

   currentAppID = 1001;
   
   Start();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Start(void)
{
   SimulationMenuStart->Enabled = false;
   SimulationMenuStart->Caption = "&Stop";

   FileMenuOpen->Enabled = false;

   // Clear messages in window
   MsgWindow->msgMemo->Clear();

   // Indicate load status
   StatusMemo->Lines->Strings[0] = "Loading application...";

   // Change offHook status
   onHook = true;

   lastStatus = lsNotRunning;

   applicationsManager->StartSimulation(currentAppID);

   SimulationMenuStart->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::End(void)
{
   SimulationMenuStart->Enabled = false;

   // Shutdown CallSystem if still running
   SetEvent(callSystemParams.shutdown);

   // Shut down engine thread
   applicationsManager->StopApplication(currentAppID);

   // Change offHook status
   onHook = true;

   ResetGraphics();

   StatusLabel->Caption = "Waiting for previous application to shut down";
}
//---------------------------------------------------------------------------

// Over-rides Windows message dispatching for TFSimulator specific events
void __fastcall TFormSimulator::Dispatch(void *Message)
{
	switch(((PMessage)Message)->Msg)
	{
		case MM_MCINOTIFY:
     	{
     		ProcessMCINotify(*((TMessage *)Message));
         break;
  	   }

      case MM_WIM_DATA:
      {
         ProcessWIMData(*((TMessage *)Message));
         break;
      }

		default:
     	{
         TForm::Dispatch(Message);
     	}
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ProcessMCINotify(TMessage& message)
{
   if((MCIDEVICEID) message.LParam == mciDevice)
  	{
      SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_PLAYRECORD_DONE);
  	}
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ProcessWIMData(TMessage& message)
{
   SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_AUDIO_CHUNK_FINISHED, message.LParam);
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::FileMenuOpenClick(TObject *Sender)
{
   if(OpenDialog->Execute())
   {
      PrepareNewApplication();
      currentApplication->info.pszAppFile = new char[OpenDialog->FileName.Length() + 1];
      strcpy(currentApplication->info.pszAppFile, OpenDialog->FileName.c_str());
      for(currentApplication->info.iParamCount = 0; currentApplication->info.iParamCount < MAX_PARAMETERS;
         currentApplication->info.iParamCount++)
      {
         currentApplication->info.pszParams[currentApplication->info.iParamCount] = new char[1];
         currentApplication->info.pszParams[currentApplication->info.iParamCount][0] = '\0';
      }

      Start();
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::SimulationMenuStartClick(TObject *Sender)
{
   if(SimulationMenuStart->Caption == "&Start")
   {
      if((currentApplication->info.pszAppFile == NULL) || (currentApplication->info.pszAppFile[0] == '\0'))
      {
         FileMenuOpenClick(Sender);
      }

      if((currentApplication->info.pszAppFile != NULL) && (currentApplication->info.pszAppFile[0] != '\0'))
      {
         Start();
      }
   }
   else
   {
      End();
   }
}
//---------------------------------------------------------------------------
// Updates the TFSimulator with new TFEngine information
void __fastcall TFormSimulator::ProcessRefresh(char status, int stepNumber, const char* stepName, const char* TAMName)
{
   char number[17];

   if((status != lsCrashed) || !LogMenuFreezeDisplayAtError->Checked)
   {
      try
      {
         // Change TAM name
         StatusGroupBox->Caption = TAMName != NULL ? TAMName : "";
      }
      catch(EInvalidOperation &discard)
      {
         // Windows occassionally throws this exception with a reason of
         // 'Canvas does not allow drawing' (which is not exactly true).
         // Prior and subsequent executions of the above statement succeed.
         // Therefore, we catch the exception and ignore it, and Simulator
         // does not crash because of it.
      }
      // Change step number and name
      if(stepNumber != -1)
      {
         sprintf(number, "[%d] ", stepNumber);
         StatusMemo->Lines->Strings[0] = number;
         StatusMemo->Lines->Strings[0] = StatusMemo->Lines->Strings[0] + stepName;
      }
   }

   // Change status light
   if(status != lastStatus)
   {
      NormalStatusLightImage->Hide();
      CrashedStatusLightImage->Hide();
      NotRunningStatusLightImage->Hide();

      switch(status)
      {
         case lsNormal:
         {
            StatusLabel->Caption = "Running normally";
            NormalStatusLightImage->Show();
            break;
         }

         case lsWarning:
         {
            StatusLabel->Caption = "Running normally after restarting";
            break;
         }

         case lsCrashed:
         {
            StatusLabel->Caption = "Terminated abnormally";
            CrashedStatusLightImage->Show();
            SimulationMenuStart->Enabled = false;
            break;
         }

         case lsNotRunning:
         {
            StatusLabel->Caption = "Completed normally";
            NotRunningStatusLightImage->Show();
            break;
         }

         case lsEmpty:
         default:
         {
            StatusLabel->Caption = "";
         }
      }

      lastStatus = status;
   }

   if(PauseDisabledImage->Visible)
   {
      PauseDisabledImage->Hide();
      PauseImage->Show();
      SimulationMenuPause->Enabled = true;
   }

   if(ResumeDisabledImage->Visible)
   {
      ResumeDisabledImage->Hide();
      ResumeImage->Show();
      SimulationMenuPause->Enabled = true;
   }

   if(NextStepDisabledImage->Visible)
   {
      NextStepDisabledImage->Hide();
      NextStepImage->Show();
   }
}

void __fastcall TFormSimulator::ProcessOnHook(void)
{
   if(!onHook)
   {
      // If TFSimulator calling, end place call thread
      if(callSystemID != 0)
      {
         // Set call system shutdown event
         SetEvent(callSystemParams.shutdown);
         callSystemID = 0;
      }
      // If in a call, play remote disconnect noise and go on hook
      else if(connected)
      {
         PlaySound("HANGUP", HInstance, SND_RESOURCE | SND_ASYNC);
      }
   }

   ResetGraphics();
}

void __fastcall TFormSimulator::ProcessOffHook(void)
{
   // If TFSimulator calling, end place call thread
   if(callSystemID != 0)
   {
      // Set call system shutdown event
      SetEvent(callSystemParams.shutdown);
      callSystemID = 0;
   }

   connected = true;
   onHook = false;
   OnHookImage->Hide();
   OffHookImage->Show();
   HangupImage->Show();
}

void __fastcall TFormSimulator::ProcessReadyForCall(bool on)
{
   readyForCall = on;
   CallSystemDisabledImage->Visible = !on;
   CallSystemImage->Visible = on;
   CallSystemDownImage->Hide();
}

void __fastcall TFormSimulator::ProcessOutgoingCall(void)
{
   PlaceCallForm->GetResponse(this);
}

void __fastcall TFormSimulator::SetHWI(HWAVEIN hwi)
{
   this->hwi = hwi;
}

void __fastcall TFormSimulator::SetMCIDevice(LONG mciDevice)
{
   this->mciDevice = mciDevice;
}

void __fastcall TFormSimulator::ShowStopRecordingButton(bool on)
{
   StopRecordingImage->Visible = on;
   StopRecordingDownImage->Hide();
   StopRecordingDisabledImage->Hide();
}

void __fastcall TFormSimulator::ShowBargeInButton(bool on)
{
   BargeInImage->Hide();
   BargeInDownImage->Hide();
   BargeInDisabledImage->Visible = on;
}

void __fastcall TFormSimulator::EnableBargeInButton(bool enabled)
{
   BargeInImage->Visible = enabled;
   BargeInDownImage->Hide();
   BargeInDisabledImage->Visible = !enabled;
}

// Called by MessageMonitor when TFEngine has sent message that it has finished
void __fastcall TFormSimulator::TerminateEngine(int appID)
{
   if((currentApplication != NULL) && (currentApplication->info.iAppID == appID))
   {
      // Shutdown CallSystem if still running
      SetEvent(callSystemParams.shutdown);

      SimulationMenuStart->Enabled = true;
      SimulationMenuStart->Caption = "&Start";

      FileMenuOpen->Enabled = true;
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::LogMenuMaximumShortPlayClick(TObject *Sender)
{
   LogMenuMaximumShortPlay->Checked = !LogMenuMaximumShortPlay->Checked;
   SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_CHANGE_SHORT_AUDIO_LENGTH, LogMenuMaximumShortPlay->Checked ? 2000 : 0);
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::FileMenuExitClick(TObject *Sender)
{
   Close();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
   TShiftState MouseShift;
   MouseShift << ssLeft;

   switch(Key)
   {
      case VK_NUMPAD1:
      case '1': Button1ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD2:
      case '2': Button2ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD3: Button3ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case '3': Shift.Contains(ssShift) ? ButtonPoundImageMouseDown(Sender, mbLeft, MouseShift, 28, 18)
          : Button3ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD4:
      case '4': Button4ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD5:
      case '5': Button5ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD6:
      case '6': Button6ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD7:
      case '7': Button7ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD8: Button8ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case '8': Shift.Contains(ssShift) ? ButtonStarImageMouseDown(Sender, mbLeft, MouseShift, 28, 18)
          : Button8ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD9:
      case '9': Button9ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD0:
      case '0': Button0ImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_MULTIPLY:
      case 's': ButtonStarImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_DIVIDE:
      case 'p': ButtonPoundImageMouseDown(Sender, mbLeft, MouseShift, 28, 18); break;
   }
}
//---------------------------------------------------------------------------
void __fastcall TFormSimulator::FormKeyUp(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
   TShiftState MouseShift;
   MouseShift << ssLeft;

   switch(Key)
   {
      case VK_NUMPAD1:
      case '1': Button1ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD2:
      case '2': Button2ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD3: Button3ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case '3': Shift.Contains(ssShift) ? ButtonPoundImageMouseUp(Sender, mbLeft, MouseShift, 28, 18)
          : Button3ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD4:
      case '4': Button4ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD5:
      case '5': Button5ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD6:
      case '6': Button6ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD7:
      case '7': Button7ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD8: Button8ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case '8': Shift.Contains(ssShift) ? ButtonStarImageMouseUp(Sender, mbLeft, MouseShift, 28, 18)
          : Button8ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD9:
      case '9': Button9ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_NUMPAD0:
      case '0': Button0ImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_MULTIPLY:
      case 's': ButtonStarImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
      case VK_DIVIDE:
      case 'p': ButtonPoundImageMouseUp(Sender, mbLeft, MouseShift, 28, 18); break;
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::AddLine(const char* line)
{
   string memoLine;
   string::size_type startPos = 0;

   if((line == NULL) || (MsgWindow->msgMemo == NULL))
   {
      return;
   }

   memoLine = line;

   // Separate lines
   for(string::size_type endPos = memoLine.find('\n', startPos); endPos != string::npos;
      endPos = memoLine.find('\n', startPos))
   {
      MsgWindow->msgMemo->Lines->Add(memoLine.substr(startPos, endPos).c_str());
      startPos += endPos + 1;
   }

   MsgWindow->msgMemo->Lines->Add(memoLine.substr(startPos).c_str());
}
//---------------------------------------------------------------------------

SimLog::SimLog(TFormSimulator* simulator, int maxFileSize, unsigned int maxLines, unsigned int maxLineSize)
   : GUILog(maxFileSize, maxLines, maxLineSize), simulator(simulator)
{
}
//---------------------------------------------------------------------------

SimLog::~SimLog(void)
{
}

void SimLog::PrintLine(const std::string& asLine, bool)
{
   mutex.Get();

   if(simulator != NULL)
   {
      simulator->AddLine(asLine.c_str());
   }

   mutex.Release();
}
//---------------------------------------------------------------------------

void SimLog::Deactivate(void)
{
   mutex.Get();

   simulator = NULL;

   mutex.Release();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::LogMenuFreezeDisplayAtErrorClick(
      TObject *Sender)
{
   LogMenuFreezeDisplayAtError->Checked = !LogMenuFreezeDisplayAtError->Checked;
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::FileMenuSetLogDirectoryClick(
      TObject *Sender)
{
   LogDirForm->SetDirectory(regInfo.pszLogDir != NULL ? regInfo.pszLogDir : "");

   if(LogDirForm->ShowModal() == mrOk)
   {
      delete[] regInfo.pszLogDir;
      regInfo.pszLogDir = new char[LogDirForm->GetDirectory().Length() + 1];
      strcpy(regInfo.pszLogDir, LogDirForm->GetDirectory().c_str());
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::LogMenuLogDirectoryClick(TObject *Sender)
{
   FileMenuSetLogDirectoryClick(Sender);
}
//---------------------------------------------------------------------------

/* -------------------------------------------------------------------------
   Name:        TFormMon::ReadTFSimulatorRegistry

   Purpose:     Reads in the registry values into application memory.

   Notes:       - When a new field is added to the form, it won't exist in
                  the registry.  Don't stop loading registry settings if one
                  value is missing.
   ------------------------------------------------------------------------- */
void __fastcall TFormSimulator::ReadTFSimulatorRegistry(void)
{
   HKEY  hkey;
   DWORD dwLen;
   DWORD dwValue;
   DWORD dwType;
   char  szKey[30];  // Capable of holding "Host %d" for up to 64-bit int counter
   time_t systime;
   struct tm *time_now;

   // Set "TFSTART" in regInfo object, but do not write to registry (this is a simulation only)
   // Indicates last date/time TFServer started
   time(&systime);
   time_now = localtime(&systime);
   regInfo.pszTFStart = new char[15];
   strftime((char*) regInfo.pszTFStart, 15, "%Y%m%d%H%M%S", time_now);
   regInfo.pszTFStart[14] = '\0';

   try
   {
   	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, TFSIMULATOR_REG_KEY, NULL, KEY_READ, &hkey) != ERROR_SUCCESS)
      {
      	return;
      }

      // Get length of string
      if(RegQueryValueEx(hkey, "LogDirectory", NULL, &dwType, NULL, &dwLen) == ERROR_SUCCESS)
      {
         // Delete directory specified for TFServer in regInfo
         delete[] regInfo.pszLogDir;
      	// Create copy of TFSimulator log directory
         regInfo.pszLogDir = new char[dwLen + 1];
         RegQueryValueEx(hkey, "LogDirectory", NULL, &dwType, regInfo.pszLogDir, &dwLen);
      }

      if(RegQueryValueEx(hkey, "LogLines", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	regInfo.dwLogLines = dwValue;
      }

      if(RegQueryValueEx(hkey, "ShortAudioPrompts", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	shortAudioPrompts = dwValue;
         LogMenuMaximumShortPlay->Checked = (dwValue != 0);
      }
      // Default value if registry key doesn't exists or can't be read
      else
      {
         shortAudioPrompts = 0;
         LogMenuMaximumShortPlay->Checked = false;
      }

      // Get length of string
      if(RegQueryValueEx(hkey, "FreezeDisplayAtError", NULL, &dwType, NULL, &dwLen) == ERROR_SUCCESS)
      {
      	// Create copy of string
         char* pszValue = new char[dwLen + 1];

         if(RegQueryValueEx(hkey, "FreezeDisplayAtError", NULL, &dwType, pszValue, &dwLen) == ERROR_SUCCESS)
         {
         	LogMenuFreezeDisplayAtError->Checked = !strcmp(pszValue, "True");
         }

         delete[] pszValue;
      }
      // Default value if registry key doesn't exists or can't be read
      else
      {
      	LogMenuFreezeDisplayAtError->Checked = true;
      }

      if(RegQueryValueEx(hkey, "Top", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	Top = dwValue;
      }

      if(RegQueryValueEx(hkey, "Left", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	Left = dwValue;
      }

      if(RegQueryValueEx(hkey, "MessagesTop", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	MsgWindow->Top = dwValue;
      }

      if(RegQueryValueEx(hkey, "MessagesLeft", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	MsgWindow->Left = dwValue;
      }

      if(RegQueryValueEx(hkey, "MessagesHeight", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	MsgWindow->Height = dwValue;
      }

      if(RegQueryValueEx(hkey, "MessagesWidth", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	MsgWindow->Width = dwValue;
      }

      if(RegQueryValueEx(hkey, "MessagesState", NULL, &dwType, (LPBYTE) &dwValue, &dwLen) == ERROR_SUCCESS)
      {
      	MsgWindow->WindowState = (TWindowState) dwValue;
      }

      RegCloseKey(hkey);
   }
   catch(Exception &e)
   {
      Application->MessageBox(e.Message.c_str(), "ReadTFSimulatorRegistry", MB_OK | MB_ICONERROR);
   }
}
//---------------------------------------------------------------------------

/* -------------------------------------------------------------------------
   Name:        TFormMon::WriteRegistry
   ------------------------------------------------------------------------- */
void __fastcall TFormSimulator::WriteRegistry(void)
{
   LONG  ret;
   HKEY  hkey;
   DWORD dwDisposition;
   DWORD dwValue;
   DWORD dwLen;
   char  szKey[30];  // Capable of holding "Host %d" for up to 64-bit int counter
   char szValue[255];

   try
   {
      ret = RegCreateKeyEx(
         HKEY_LOCAL_MACHINE,      // Handle of an open key.
         TFSIMULATOR_REG_KEY,     // Address of subkey name.
         0,                       // Reserved.
         NULL,                    // Address of class string.
         REG_OPTION_NON_VOLATILE, // Special options flag.
         KEY_WRITE,               // Desired security access.
         NULL,                    // Address of key security structure.
         &hkey,                   // Address of buffer for opened handle.
         &dwDisposition);         // Address of disposition value buffer.

      if(ret != ERROR_SUCCESS)
      {
         Application->MessageBox("Failed to create registry key.", "Registry Save Failed",
            MB_ICONERROR | MB_OK);
      }
      else
      {
      	ret = RegSetValueEx(hkey, "LogDirectory", 0, REG_SZ, regInfo.pszLogDir, strlen(regInfo.pszLogDir) + 1);

         ret &= RegSetValueEx(hkey, "LogLines", 0, REG_DWORD, (LPBYTE) &regInfo.dwLogLines, sizeof(DWORD));

         ret &= RegSetValueEx(hkey, "ShortAudioPrompts", 0, REG_DWORD, (LPBYTE) &shortAudioPrompts, sizeof(DWORD));

         ret &= RegSetValueEx(hkey, "FreezeDisplayAtError", 0, REG_SZ, LogMenuFreezeDisplayAtError->Checked ? "True" : "False", LogMenuFreezeDisplayAtError->Checked ? 5 : 6);

         ret &= RegSetValueEx(hkey, "Top", 0, REG_DWORD, (LPBYTE) &Top, sizeof(DWORD));

         ret &= RegSetValueEx(hkey, "Left", 0, REG_DWORD, (LPBYTE) &Left, sizeof(DWORD));

         ret &= RegSetValueEx(hkey, "MessagesTop", 0, REG_DWORD, (LPBYTE) &(MsgWindow->Top), sizeof(DWORD));

         ret &= RegSetValueEx(hkey, "MessagesLeft", 0, REG_DWORD, (LPBYTE) &(MsgWindow->Left), sizeof(DWORD));

         ret &= RegSetValueEx(hkey, "MessagesHeight", 0, REG_DWORD, (LPBYTE) &(MsgWindow->Height), sizeof(DWORD));

         ret &= RegSetValueEx(hkey, "MessagesWidth", 0, REG_DWORD, (LPBYTE) &(MsgWindow->Width), sizeof(DWORD));

         ret &= RegSetValueEx(hkey, "MessagesState", 0, REG_DWORD, (LPBYTE) &(MsgWindow->WindowState), sizeof(DWORD));

         if(ret != ERROR_SUCCESS)
         {
            Application->MessageBox("Failed to write host entry.", "Registry Save Failed",
               MB_ICONERROR | MB_OK);
         }
      }

      RegCloseKey(hkey);
   }
   catch(Exception &e)
   {
      Application->MessageBox(e.Message.c_str(), "WriteRegistry", MB_OK | MB_ICONERROR);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::HelpMenuHelpClick(TObject *Sender)
{
   // function in ESCUTIL: help_fns.cpp, opens help file and links to topic requested
   HelpLink("SimulatorCHAP.htm");
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::HelpMenuAboutClick(TObject *Sender)
{
   AboutBox->ShowModal();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button1ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button1DownImage->Show();
      SendDigit('1');
      PlaySound("TOUCHTONE1", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button1ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button1DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button2ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button2DownImage->Show();
      SendDigit('2');
      PlaySound("TOUCHTONE2", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button2ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button2DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button3ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button3DownImage->Show();
      SendDigit('3');
      PlaySound("TOUCHTONE3", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button3ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button3DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button4ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button4DownImage->Show();
      SendDigit('4');
      PlaySound("TOUCHTONE4", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button4ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button4DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button5ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button5DownImage->Show();
      SendDigit('5');
      PlaySound("TOUCHTONE5", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button5ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button5DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button6ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button6DownImage->Show();
      SendDigit('6');
      PlaySound("TOUCHTONE6", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button6ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button6DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button7ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button7DownImage->Show();
      SendDigit('7');
      PlaySound("TOUCHTONE7", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button7ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button7DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button8ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button8DownImage->Show();
      SendDigit('8');
      PlaySound("TOUCHTONE8", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button8ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button8DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button9ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button9DownImage->Show();
      SendDigit('9');
      PlaySound("TOUCHTONE9", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button9ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button9DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ButtonStarImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      ButtonStarDownImage->Show();
      SendDigit('*');
      PlaySound("TOUCHTONESTAR", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ButtonStarImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   ButtonStarDownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button0ImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      Button0DownImage->Show();
      SendDigit('0');
      PlaySound("TOUCHTONE0", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Button0ImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   Button0DownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ButtonPoundImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(!onHook && connected)
   {
      ButtonPoundDownImage->Show();
      SendDigit('#');
      PlaySound("TOUCHTONEPOUND", HInstance, SND_RESOURCE | SND_ASYNC);
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ButtonPoundImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   ButtonPoundDownImage->Hide();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Pause(void)
{
   if(currentApplication != NULL)
   {
      ResetEvent(currentApplication->info.hResume);
      PauseImage->Hide();
      PauseDownImage->Hide();
      ResumeDisabledImage->Show();
      NextStepDisabledImage->Show();

      SimulationMenuPause->Caption = "&Resume";
      SimulationMenuPause->Enabled = false;
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::PauseImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   PauseImage->Hide();
   PauseDownImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::PauseImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(PauseDownImage->Visible)
   {
      Pause();
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ResumeImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   ResumeImage->Hide();
   ResumeDownImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::Resume(void)
{
   if(currentApplication != NULL)
   {
      SetEvent(currentApplication->info.hResume);
      ResumeImage->Hide();
      ResumeDownImage->Hide();
      PauseImage->Show();
      NextStepImage->Hide();

      SimulationMenuPause->Caption = "&Pause";
      SimulationMenuPause->Enabled = true;
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ResumeImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(ResumeDownImage->Visible)
   {
      Resume();
   }
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ShowLogImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   ShowLogImage->Hide();
   ShowLogDownImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::ShowLogImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   ShowLogDownImage->Hide();
   ShowLogImage->Show();

	// Show the window
	MsgWindow->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::StopRecordingImageMouseDown(
      TObject *Sender, TMouseButton Button, TShiftState Shift, int X,
      int Y)
{
   StopRecordingImage->Hide();
   StopRecordingDownImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::StopRecordingImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_INTERRUPTPLAYRECORD);
   StopRecordingDownImage->Hide();
}
//---------------------------------------------------------------------------


void __fastcall TFormSimulator::CallSystemImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   CallSystemImage->Hide();
   CallSystemDownImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::CallSystemImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   CallSystemDownImage->Hide();

   // We are calling the board
   if(readyForCall)
   {
      // Initialise call system parameters
      ResetEvent(callSystemParams.shutdown);

      // Start thread to call system until cancel or pickup occurs
      _beginthreadex(
         NULL,               // NT security attributes
         0,                  // stack size (0 = use same size as this thread)
         CallSystem,         // start address
         &callSystemParams,  // argument list
         0,                  // Thread creation flags
         &callSystemID);     // Thread id
   }

   // Change offHook status
   onHook = false;
   OnHookImage->Hide();
   OffHookImage->Show();

   // Change Image
   HangupImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::HangupImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   HangupImage->Hide();
   HangupDownImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::HangupImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   HangupDownImage->Hide();

   // Calling cancelled
   if(callSystemID != 0)
   {
      // Set call system shutdown event
      SetEvent(callSystemParams.shutdown);
      callSystemID = 0;
   }

   if(connected)
   {
      // Send hangup event
      SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_ON_HOOK);
   }

   // Change offHook status
   onHook = true;
   OnHookImage->Show();
   OffHookImage->Hide();

   // System still waiting for call
   if(readyForCall)
   {
      CallSystemImage->Show();
   }
   else
   // System not waiting for call
   {
      CallSystemDisabledImage->Show();
   }

   PlaySound("HANGUP", HInstance, SND_RESOURCE | SND_ASYNC);
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::NextStepImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   NextStepImage->Hide();
   NextStepDownImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::NextStepImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(currentApplication != NULL)
   {
      PulseEvent(currentApplication->info.hResume);
      NextStepDownImage->Hide();
      NextStepDisabledImage->Show();
      ResumeImage->Hide();
      ResumeDisabledImage->Show();
   }
}
//---------------------------------------------------------------------------


void __fastcall TFormSimulator::SimulationMenuPauseClick(TObject *Sender)
{
   (SimulationMenuPause->Caption == "&Pause") ? Pause() : Resume();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_EVENT event, int eventInfo)
{
   TFEngineCommunicationPacket* engineMessage = new TFEngineCommunicationPacket(currentAppID, TFEngineCommunicationPacket::SRV_EVT_BOARD_EVENT);
   BoardSimulatorCommunicationPacket* boardMessage = new BoardSimulatorCommunicationPacket(event);
   boardMessage->eventInfo = eventInfo;
   engineMessage->boardMessage = boardMessage;
   commHub.SendMessage(TFServerCommunicationPacket::TF_ENGINE_MANAGER, engineMessage);
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::SendDigit(char digit)
{
   SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_DIGIT, digit);
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::BargeInImageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   BargeInImage->Hide();
   BargeInDownImage->Show();
}
//---------------------------------------------------------------------------

void __fastcall TFormSimulator::BargeInImageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   BargeInDownImage->Hide();
   SendEvent(BoardSimulatorCommunicationPacket::BRD_SIMULATOR_BARGEIN);
}
//---------------------------------------------------------------------------

